@@ -1,5 +1,5 @@
 package Devel::Camelcadedb;
-our $VERSION = "1.6.1.1";
+our $VERSION = "1.6.1.2";
 
 # http://perldoc.perl.org/DB.html
 # http://perldoc.perl.org/perldebug.html
@@ -10,10 +10,12 @@ use 5.008;
 use strict;
 use warnings;
 use IO::Socket::INET;
+use IO::Select;
 use PadWalker qw/peek_my peek_our/;
 use Scalar::Util;
 use Encode;
 use overload;
+#use Carp;
 
 #sub FLAG_REPORT_GOTO() {0x80;}
 
@@ -118,6 +120,7 @@ my $_skip_run_stop = 0; # flag for skipping forced stop on run phase
 
 my $_debug_socket;
 my $_debug_packed_address;
+my IO::Select $_debug_socket_select;
 
 my $coder;  # JSON::XS coder
 my $deparser; # B::Deparse deparser
@@ -674,7 +677,7 @@ sub _render_variables
 
 sub _get_current_stack_frame
 {
-    return $_stack_frames->[0];
+    return $_stack_frames->[-1];
 }
 
 sub _send_data_to_debugger
@@ -791,13 +794,144 @@ sub _is_use_frame
     return $is_use_block;
 }
 
-sub _release_the_hounds
+sub _set_frames_single
 {
+    my ($new_value) = @_;
     foreach my $frame (@{$_stack_frames})
     {
-        $frame->{single} = STEP_CONTINUE;
+        $frame->{single} = $new_value;
     }
-    $DB::single = STEP_CONTINUE;
+    $DB::single = $new_value;
+}
+
+sub _hold_the_line
+{
+    _set_frames_single( STEP_INTO );
+}
+
+sub _release_the_hounds
+{
+    _set_frames_single( STEP_CONTINUE );
+}
+
+sub _process_command
+{
+    my ($command) = @_;
+    _report "============> Got command: '%s'\n", $command if $_dev_mode;
+
+    if ($command eq 'q')
+    {
+        _report "Exiting" if $_dev_mode;
+        exit;
+    }
+    elsif ($command =~ /^e\s+(.+)$/) # eval expresion
+    {
+        my $data = $1;
+        my $transaction_data = _deserialize( $data );
+
+        my ($transaction_id, $request_object) = @$transaction_data{qw/id data/};
+
+        my $result = _eval_expression( _from_utf8( $request_object->{expression} // '' ) );
+        $result->{result} = _get_reference_descriptor( result => $result->{result} );
+
+        _send_transaction_response( $transaction_id, $result );
+
+        _report "Result is $result\n" if $_dev_mode;
+    }
+    elsif ($command eq 'pause')
+    {
+        _hold_the_line;
+    }
+    elsif ($command eq 'g')
+    {
+        _release_the_hounds();
+        return;
+    }
+    elsif ($command =~ /^b (.+)$/) # set breakpoints from proto
+    {
+        _process_new_breakpoints( $1 );
+    }
+    elsif ($command =~ /^p (.+)$/) # Run to cursor
+    {
+        _set_run_to_cursor_breakpoint( $1 );
+        _release_the_hounds();
+        return;
+    }
+    elsif ($command eq 'o') # over,
+    {
+        my $current_frame = _get_current_stack_frame;
+        if (_is_use_frame( $current_frame ))
+        {
+            $current_frame->{single} = STEP_INTO;
+            $DB::single = STEP_CONTINUE;
+        }
+        else
+        {
+            $DB::single = STEP_OVER;
+        }
+        return;
+    }
+    elsif ($command =~ /^getchildren (.+)$/) # expand,
+    {
+        _get_reference_subelements( $1 );
+    }
+    elsif ($command =~ /^get_source (.+)$/) # get eval/file source
+    {
+        _get_file_source_handler( $1 );
+    }
+    elsif ($command eq 'u') # step out
+    {
+        my $current_frame = _get_current_stack_frame;
+        if (_is_use_frame( $current_frame ))
+        {
+            $current_frame->{single} = STEP_CONTINUE;
+        }
+        $DB::single = STEP_CONTINUE;
+
+        return;
+    }
+    else
+    {
+        $DB::single = STEP_INTO;
+        return;
+    }
+    return 1;
+}
+
+my $input_buffer = '';
+
+sub _get_next_command
+{
+    my $read_bytes;
+    my $new_line_index = index $input_buffer, "\n";
+
+    if ($new_line_index == -1)
+    {
+        while( $read_bytes = sysread( $_debug_socket, $input_buffer, 10240, length( $input_buffer ) ))
+        {
+            last if ($new_line_index = index $input_buffer, "\n") > -1;
+        }
+        unless (defined $read_bytes)
+        {
+            die 'Debugging socket disconnected';
+            print STDERR "$!\n";
+            exit -1;
+        }
+        unless ($new_line_index > -1)
+        {
+            print STDERR "Buffer $input_buffer has no newlines in it and nothing is in the socket\n";
+            exit -1;
+        }
+    }
+    my $command = substr $input_buffer, 0, $new_line_index + 1, '';
+    $command =~ s/[\r\n]+$//;
+    #    printf STDERR "Got command: %s and left in buffer %s\n%s\n\%s\n", length $command, length $input_buffer ,$command, $input_buffer;
+    return $command;
+}
+
+sub _can_read
+{
+    return length( $input_buffer ) > 0 || ($_debug_socket && $_debug_socket_select && scalar $_debug_socket_select->can_read( 0 ));
 }
 
 sub _event_handler
@@ -817,84 +951,7 @@ sub _event_handler
     while()
     {
         _report "Waiting for input\n" if $_dev_mode;
-        local $/ = "\n";
-        my $command = <$_debug_socket>;
-        die 'Debugging socket disconnected' if !defined $command;
-        $command =~ s/[\r\n]+$//;
-        _report "============> Got command: '%s'\n", $command if $_dev_mode;
-
-        if ($command eq 'q')
-        {
-            _report "Exiting" if $_dev_mode;
-            exit;
-        }
-        elsif ($command =~ /^e\s+(.+)$/) # eval expresion
-        {
-            my $data = $1;
-            my $transaction_data = _deserialize( $data );
-
-            my ($transaction_id, $request_object) = @$transaction_data{qw/id data/};
-
-            my $result = _eval_expression( _from_utf8( $request_object->{expression} // '' ) );
-            $result->{result} = _get_reference_descriptor( result => $result->{result} );
-
-            _send_transaction_response( $transaction_id, $result );
-
-            _report "Result is $result\n" if $_dev_mode;
-        }
-        elsif ($command eq 'g')
-        {
-            _release_the_hounds();
-            return;
-        }
-        elsif ($command =~ /^b (.+)$/) # set breakpoints from proto
-        {
-            _process_new_breakpoints( $1 );
-        }
-        elsif ($command =~ /^p (.+)$/) # Run to cursor
-        {
-            _set_run_to_cursor_breakpoint( $1 );
-            _release_the_hounds();
-            return;
-        }
-        elsif ($command eq 'o') # over,
-        {
-            my $current_frame = _get_current_stack_frame;
-            if (_is_use_frame( $current_frame ))
-            {
-                $current_frame->{single} = STEP_INTO;
-                $DB::single = STEP_CONTINUE;
-            }
-            else
-            {
-                $DB::single = STEP_OVER;
-            }
-            return;
-        }
-        elsif ($command =~ /^getchildren (.+)$/) # expand,
-        {
-            _get_reference_subelements( $1 );
-        }
-        elsif ($command =~ /^get_source (.+)$/) # get eval/file source
-        {
-            _get_file_source_handler( $1 );
-        }
-        elsif ($command eq 'u') # step out
-        {
-            my $current_frame = _get_current_stack_frame;
-            if (_is_use_frame( $current_frame ))
-            {
-                $current_frame->{single} = STEP_CONTINUE;
-            }
-            $DB::single = STEP_CONTINUE;
-
-            return;
-        }
-        else
-        {
-            $DB::single = STEP_INTO;
-            return;
-        }
+        _process_command( _get_next_command ) || return;
     }
 }
 
@@ -917,7 +974,7 @@ sub _enter_frame
         subname => $DB::sub,
         single  => $old_db_single,
     };
-    unshift @$_stack_frames, $new_stack_frame;
+    push @{$_stack_frames}, $new_stack_frame;
     _apply_queued_breakpoints() if $ready_to_go;
     return $new_stack_frame;
 }
@@ -926,7 +983,7 @@ sub _enter_frame
 sub _exit_frame
 {
     $_internal_process = 1;
-    my $frame = shift @$_stack_frames;
+    my $frame = pop @$_stack_frames;
     $frame_prefix = $frame_prefix_step x (scalar @$_stack_frames);
     _report "Leaving frame %s, setting single to %s", (scalar @$_stack_frames + 1),
         $frame->{single} if $_debug_sub_handler && $_dev_mode;
@@ -1257,6 +1314,18 @@ sub _set_break_points_for_file
     _switch_context( $old_context );
 }
 
+#sub mydie
+#{
+#    my ($msg) = @_;
+#    print "$msg\n";
+#    print Carp::longmess;
+#    foreach my $key (sort keys %::)
+#    {
+#        print $key."\n" if $key =~ /^_</;
+#    }
+#    exit -1;
+#}
+
 sub _calc_real_path
 {
     my $path = shift;
@@ -1270,9 +1339,9 @@ sub _calc_real_path
     else
     {
         $real_path = eval {Cwd::realpath( $path )};
-        if (my $e = $@)
+        unless ($real_path)
         {
-            _report "Error on getting real path for $path, $e" if $_dev_mode;
+            _report 'Unable to find real path for %s use as it is', $path if $_dev_mode;
             $real_path = $path;
         }
         $real_path =~ s{\\}{/}g;
@@ -1469,6 +1538,9 @@ sub sub_handler
     {
         $_internal_process = 1;
 
+        _process_command( _get_next_command ) while _can_read;
+        $old_db_single = $DB::single; # might be overriden in commands
+
         $DB::single = STEP_CONTINUE;
 
         $stack_frame = _enter_frame( $old_db_single );
@@ -1484,12 +1556,22 @@ sub sub_handler
         $_internal_process = 0;
     }
 
+    my $stack_pointer = $#$_stack_frames;
+
     my $wantarray = wantarray;
 
     if ($DB::single == STEP_OVER)
     {
-        _report "Disabling step in in subcalls\n" if $_debug_sub_handler && $_dev_mode;
+        _report "Disabling step in in subcalls, will restore %s\n",
+            $_stack_frames->[-1]->{single}
+            if $_debug_sub_handler && $_dev_mode;
         $DB::single = STEP_CONTINUE;
+
+        #        my $die_handler = $SIG{__DIE__};
+        #        local $SIG{__DIE__} = sub{
+        #            _hold_the_line;
+        #            goto &$die_handler if $die_handler;
+        #        };
     }
     else
     {
@@ -1500,7 +1582,7 @@ sub sub_handler
     {
         no strict 'refs';
         &$DB::sub;
-
+        $#$_stack_frames = $stack_pointer;
         if ($stack_frame)
         {
             _exit_frame();
@@ -1516,6 +1598,7 @@ sub sub_handler
     {
         no strict 'refs';
         my @result = &$DB::sub;
+        $#$_stack_frames = $stack_pointer;
         if ($stack_frame)
         {
             _exit_frame();
@@ -1530,6 +1613,7 @@ sub sub_handler
     {
         no strict 'refs';
         my $result = &$DB::sub;
+        $#$_stack_frames = $stack_pointer;
         if ($stack_frame)
         {
             _exit_frame();
@@ -1690,6 +1774,9 @@ else
     die "Error connecting to $ENV{PERL5_DEBUG_HOST}:$ENV{PERL5_DEBUG_PORT}" unless $_debug_socket;
 }
 $_debug_socket->autoflush( 1 );
+$_debug_socket_select = IO::Select->new();
+$_debug_socket_select->add( $_debug_socket );
+
 print STDERR "Connected.\n";
 
 push @$_stack_frames, {
